// pnoise.h
// Version 1.2.1
// 8 Dec 2002
// Noise for procedural texture
// Based on work of K. Perlin
// See documentation
//   http://www.cs.uaf.edu/~cs381/docs/pnoise-doc.html
// for more information.
//
// Copyright 2002, 2003 Glenn G. Chappell
// The package is intended for use in the CS 381
// (Computer Graphics) class at the U. of Alaska Fairbanks.
// Anyone desiring to use this package in some other context
// should contact the author: chappellg@member.ams.org

/*
Functions are provided to generate 3-D noise of various sorts. Noise
is generated using the C standard library rand() function, which should
be seeded appropriately before calling functions in this package.

The main functions of interest to the user are pnoise3d, fnoise3d, and
filterfnoise3d. Also, normalize3d is provided as a utility.

pnoise3d, fnoise3d, and filterfnoise3d each take, as their first
parameter, a pointer to an array of doubles. This should generally be a
3-D array. The next 3 parameters are the dimensions of this array. After
that come the "frequencies" of the noise, in the x, y, and z directions.
So, for example, if the x frequency is 10, then then 10 "cycles" of
noise will be generated in the x direction.

Noise is returned in the given array. pnoise3d returns noise in the
range [-1,1]; the average value will be zero. fnoise3d returns noise
whose range may vary; the average value will be zero. filterfnoise may
return noise in varying ranges, depending on its parameters. After
calling fnoise or filterfnoise, it is suggested that normalize3d be
called with the same array. This function normalizes the data, rescaling
it to that it has average value zero and lies in the range [-1,1].

To generate 2-D noise, use the 3-D functions, but pass a pointer to a
2-D array. The z dimension and the z frequency should both be 1.

pnoise3d returns noise similar to Ken Perlin's standard noise function.

fnoise3d returns faked 1/f noise using Ken Perlin's technique of adding
together copies of the standard noise function, repeatedly doubling the
frequency and halving the amplitude. Change the scaleratio parameter in
order to multiply the successive amplitudes by a different value.
Values of scaleratio should generally be at least 0.5 and less than 1.

filterfnoise3d passes the output of pnoise3d through a "filter" function
before summing. For example, Ken Perlin suggests adding the absolute
values of the various noise functions, in order to get noise with cusps.
The can be done by passing pnoise_abs as a filter function to
filterfnoise3d. When using filterfnoise3d, a scaleratio somewhat higher
than 0.5 is often helpful.

Suggestions on how this package may be improved are welcome!

Glenn G. Chappell
<chappellg@member.ams.org>
*/


#ifdef __sgi
# include <stdlib.h>
#else
# include <cstdlib>
# if !defined(_MSC_VER) || (_MSC_VER > 1200)
using std::rand;
# endif
#endif
#include <vector>


// pnoise_rand
// Given positive integer n.
// Returns pseudo-random integer in [0,n-1]
int pnoise_rand(int n)
{
   return rand() % n;
}


// pnoise_mod
// Given integers a, b, with b > 0, computes a mod b.
// For those who don't know what "mod" means (like the designers
//  of most modern computer languages): a mod b is nonnegative when
//  b is positive. So, for example, -1 mod 3 = 2.
// (No, I don't feel bitter about this; why do you ask?)
int pnoise_mod(int a, int b)
{
   if (a >= 0) return a % b;
   else
   {
      int c = (-a) % b;
      if (c == 0) return 0;
      else return b - c;
   }
}


// pnoise_fade
// Returns Perlin's S-shaped fade curve
// Valid in [-1,1]
double pnoise_fade(double t)
{
   double u = (t >= 0) ? 1-t : 1+t;
   return u*u*u*(u*(u*6-15)+10);
}


// pnoise_abs
// Absolute-value function, in case anyone needs one.
// This is primarily intended as a filter function for filterfnoise3d.
double pnoise_abs(double t)
{
   return (t >= 0) ? t : -t;
}


void addpnoise3d(
   double * data,                    // the 3-D array
   int dimx, int dimy, int dimz,     // array dim's
   int freqx, int freqy, int freqz,  // "frequencies"
   double scale=1.0,                 // "amplitude"
   double shiftx=0.0, double shifty=0.0, double shiftz=0.0)
                                     // phase shifts
{
   double wlenx = double(dimx) / freqx;  // "wavelengths"
   double wleny = double(dimy) / freqy;
   double wlenz = double(dimz) / freqz;

   std::vector<int> xwhichs, ywhichs, zwhichs;
   for (int wx = 0; wx < freqx; ++wx)
   {
      double centerx = wlenx * wx + shiftx;
      int lo = int(centerx - wlenx + 1);
      int hi = int(centerx + wlenx);
      if (hi >= lo)
         xwhichs.push_back(wx);
   }
   for (int wy = 0; wy < freqy; ++wy)
   {
      double centery = wleny * wy + shifty;
      int lo = int(centery - wleny + 1);
      int hi = int(centery + wleny);
      if (hi >= lo)
         ywhichs.push_back(wy);
   }
   for (int wz = 0; wz < freqz; ++wz)
   {
      double centerz = wlenz * wz + shiftz;
      int lo = int(centerz - wlenz + 1);
      int hi = int(centerz + wlenz);
      if (hi >= lo)
         zwhichs.push_back(wz);
   }

   for (std::vector<int>::const_iterator xit = xwhichs.begin(); xit != xwhichs.end(); ++xit) {
   double centerx = wlenx * (*xit) + shiftx;
   for (std::vector<int>::const_iterator yit = ywhichs.begin(); yit != ywhichs.end(); ++yit) {
   double centery = wleny * (*yit) + shifty;
   for (std::vector<int>::const_iterator zit = zwhichs.begin(); zit != zwhichs.end(); ++zit) {
   double centerz = wlenz * (*zit) + shiftz;

      int vx = 0., vy = 0., vz = 0.;
      int r = pnoise_rand(12);  // Choose random vector, len = sqrt(2)
      if (r < 8)
         vx = (r & 1) ? -1. : 1;
      else
         vy = (r & 1) ? -1. : 1;
      if (r < 4)
         vy = (r & 2) ? -1. : 1;
      else
         vz = (r & 2) ? -1. : 1;
      
      for (int ix = int(centerx-wlenx+1); ix <= int(centerx + wlenx); ++ix) {
      double tx = (ix - centerx) / wlenx;
      double fadex = pnoise_fade(tx);
      int coordx = pnoise_mod(ix, dimx);
      for (int iy = int(centery-wleny+1); iy <= int(centery + wleny); ++iy) {
      double ty = (iy - centery) / wleny;
      double fadey = pnoise_fade(ty);
      int coordy = pnoise_mod(iy, dimy);
      for (int iz = int(centerz-wlenz+1); iz <= int(centerz + wlenz); ++iz) {
      double tz = (iz - centerz) / wlenz;
      double fadez = pnoise_fade(tz);
      int coordz = pnoise_mod(iz, dimz);

         double dot = tx*vx + ty*vy + tz*vz;
         double fade = fadex * fadey * fadez;
         data[coordx*dimy*dimz + coordy*dimz + coordz] += dot * fade * scale;

      }}} // End ix/y/z for loops

   }}} // End x/y/zwhichs for loops
}


void addfnoise3d(
   double * data,                    // the 3-D array
   int dimx, int dimy, int dimz,     // array dim's
   int freqx, int freqy, int freqz,  // "frequencies"
   double scaleratio=0.5,            // scale is multiplied by this,
                                     //  when frequency is doubled
   double scale=1.0,                 // "amplitude"
   double shiftx=0.0, double shifty=0.0, double shiftz=0.0)
                                     // phase shifts
{
   int m;      // multiplier for frequency
   double sf;  // scale factor (multiplier for scale)
   for (m=1, sf=scaleratio;
        freqx*m<=dimx || freqy*m<=dimy || freqz*m<=dimz;
        m*=2, sf*=scaleratio)
   {
      addpnoise3d(
         data,
         dimx, dimy, dimz,
         m*freqx, m*freqy, m*freqz,
         sf*scale,
         shiftx+double(dimx)/(freqx*m*2),
            shifty+double(dimy)/(freqy*m*2),
            shiftz+double(dimz)/(freqz*m*2));
   }
}


void clear3d(
   double * data,                // the 3-D array
   int dimx, int dimy, int dimz) // array dim's
{
   const int size = dimx * dimy * dimz;
   for (int i=0; i<size; ++i)
      data[i] = 0.;
}


void pnoise3d(
   double * data,                    // the 3-D array
   int dimx, int dimy, int dimz,     // array dim's
   int freqx, int freqy, int freqz,  // "frequencies"
   double scale=1.0,                 // "amplitude"
   double shiftx=0.0, double shifty=0.0, double shiftz=0.0)
                                     // phase shifts
{
   clear3d(data, dimx, dimy, dimz);
   addpnoise3d(
      data,
      dimx, dimy, dimz,
      freqx, freqy, freqz,
      scale,
      shiftx, shifty, shiftz);
}


void fnoise3d(
   double * data,                    // the 3-D array
   int dimx, int dimy, int dimz,     // array dim's
   int freqx, int freqy, int freqz,  // "frequencies"
   double scaleratio=0.5,            // scale is multiplied by this,
                                     //  when frequency is doubled
   double scale=1.0,                 // "amplitude"
   double shiftx=0.0, double shifty=0.0, double shiftz=0.0)
                                     // phase shifts
{
   clear3d(data, dimx, dimy, dimz);
   addfnoise3d(
      data,
      dimx, dimy, dimz,
      freqx, freqy, freqz,
      scaleratio,
      scale,
      shiftx, shifty, shiftz);
}


void filterpnoise3d(
   double * data,                    // the 3-D array
   int dimx, int dimy, int dimz,     // array dim's
   int freqx, int freqy, int freqz,  // "frequencies"
   double (* filter)(double),        // filter function for pnoise values
   double scale=1.0,                 // "amplitude"
   double shiftx=0.0, double shifty=0.0, double shiftz=0.0)
                                     // phase shifts
{
   pnoise3d(
      data,
      dimx, dimy, dimz,
      freqx, freqy, freqz,
      scale,
      shiftx, shifty, shiftz);

   const int size = dimx * dimy * dimz;
   for (int i=0; i<size; ++i)
      data[i] = filter(data[i]);
}


void filterfnoise3d(
   double * data,                    // the 3-D array
   int dimx, int dimy, int dimz,     // array dim's
   int freqx, int freqy, int freqz,  // "frequencies"
   double (* filter)(double),        // filter function for pnoise values
   double scaleratio=0.5,            // scale is multiplied by this,
                                     //  when frequency is doubled
   double scale=1.0,                 // "amplitude"
   double shiftx=0.0, double shifty=0.0, double shiftz=0.0)
                                     // phase shifts
{
   const int size = dimx * dimy * dimz;
   double * tempdata = new double[size];

   clear3d(data, dimx, dimy, dimz);
   int m;      // multiplier for frequency
   double sf;  // scale factor (multiplier for scale)
   for (m=1, sf=scaleratio;
        freqx*m<=dimx || freqy*m<=dimy || freqz*m<=dimz;
        m*=2, sf*=scaleratio)
   {
      filterpnoise3d(
         tempdata,
         dimx, dimy, dimz,
         m*freqx, m*freqy, m*freqz,
         filter,
         sf*scale,
         shiftx+double(dimx)/(freqx*m*2),
            shifty+double(dimy)/(freqy*m*2),
            shiftz+double(dimz)/(freqz*m*2));
      for (int i=0; i<size; ++i)
         data[i] += tempdata[i];
   }

   delete [] tempdata;
}


void normalize3d(
   double * data,                // the 3-D array
   int dimx, int dimy, int dimz) // array dim's
{
   double max = data[0];
   double min = data[0];
   double sum = data[0];

   const int size = dimx * dimy * dimz;
   int i;
   for (i=1; i<size; ++i)
   {
      if (data[i] > max) max = data[i];
      if (data[i] < min) min = data[i];
      sum += data[i];
   }

   const double avg = sum/size;
   double amp = max-avg;
   if (avg-min > amp) amp = avg-min;
   if (amp == 0.) amp = 1.;

   for (i=0; i<size; ++i)
   {
      data[i] = (data[i]-avg)/amp;
   }
}

